"""
NOTE: Not all of the Samples are valid with our current implementation. The list of 
valid samples is saved in the `summoning-the-shoggoth` repo for now. Running this file
directly from the command line will run the full validation script and overwrite the
results there (takes a few hours).
"""
import argparse
import json
import os
from pathlib import Path
from typing import Any

from datasets import DatasetDict, load_dataset  # type: ignore

from benchmarks.run_sample import validate_test_fields
from mentat.sampler.sample import Sample

SWE_BENCH_SAMPLES_DIR = Path(__file__).parent / "benchmarks" / "swe_bench_samples"
SWE_VALIDATION_RESULTS_PATH = (
    Path(__file__).parent.parent.parent
    / "summoning-the-shoggoth"
    / "swe_bench"
    / "swe_bench_validation_results_2024-03-29.json"
)


def download_swe_benchmarks(split: str = "dev") -> list[dict[str, str]]:
    """Get raw SWE-Bench json samples from huggingface

    3 splits are available: dev (225), test (2.29k), and train (19k).
    """
    dataset: DatasetDict = load_dataset("princeton-nlp/SWE-bench", split=split)  # type: ignore
    dataset: list[dict[str, str]] = [dict(benchmark) for benchmark in dataset]
    return dataset


def get_swe_samples(split: str = "dev", max_benchmarks: int | None = None) -> list[Sample]:
    """Return a list of SWE-Bench samples.

    If missing, download, convert to Samples and save locally.
    """
    split_dir = SWE_BENCH_SAMPLES_DIR / split
    saved_benchmarks = list(split_dir.glob("*.json"))
    if not split_dir.exists() or max_benchmarks and len(saved_benchmarks) < max_benchmarks:
        print(f"Downloading {split} split from SWE-Bench...")
        split_dir.mkdir(parents=True, exist_ok=True)
        dataset = download_swe_benchmarks(split)
        samples = [Sample.from_swe_bench(benchmark) for benchmark in dataset]
        for sample in samples:
            sample.save(split_dir / f"{sample.id}.json")
    else:
        samples = [Sample.load(fname) for fname in saved_benchmarks]

    # Check that samples are valid
    valid_samples = list[Sample]()
    if not SWE_VALIDATION_RESULTS_PATH.exists():
        print(f"Sample validation results not found at {SWE_VALIDATION_RESULTS_PATH}.")
        print("Validating SWE samples...")
        print("\033[93m" + "Warning: This will take a couple hours." + "\033[0m")
        # This takes a couple hours.
        validate_swe_samples()
    with open(SWE_VALIDATION_RESULTS_PATH, "r") as f:
        swe_validation_results = json.load(f)
    for sample in samples:
        results = swe_validation_results.get(sample.title)
        pass_to_pass = (
            "PASS_TO_PASS" in results and results["PASS_TO_PASS"]["passed"] == results["PASS_TO_PASS"]["total"]
        )
        fail_to_pass_post = (
            "FAIL_TO_PASS_POST" in results
            and results["FAIL_TO_PASS_POST"]["passed"] == results["FAIL_TO_PASS_POST"]["total"]
        )
        if pass_to_pass and fail_to_pass_post:
            valid_samples.append(sample)
    samples = valid_samples

    if max_benchmarks:
        samples = samples[:max_benchmarks]
    print(f"Selected {len(samples)} benchmarks from '{split}'")
    return samples


def validate_swe_samples(targets: list[str] | None = None, refresh: bool = True) -> None:
    """Setup each sample and run its validation tests."""
    cwd = Path.cwd()
    samples = get_swe_samples()

    if SWE_VALIDATION_RESULTS_PATH.exists():
        with open(SWE_VALIDATION_RESULTS_PATH, "r") as f:
            results = json.load(f)
        print(f"Loaded {len(results)} previous validation results.")
    else:
        SWE_VALIDATION_RESULTS_PATH.mkdir(parents=True, exist_ok=True)
        SWE_VALIDATION_RESULTS_PATH.touch()
        results = dict[str, Any]()
    for sample in samples:
        os.chdir(cwd)

        if targets and not any(b in sample.title for b in targets):
            continue
        if not refresh and sample.title in results:
            continue
        try:
            print(80 * "*" + f"\nValidating {sample.id}...")
            test_results = validate_test_fields(sample)
            percentages = dict[str, float]()
            for category, result in test_results.items():
                _passed, _total = result.get("passed", 0), result.get("total", 0)
                percentage = 0 if _total == 0 else _passed / _total * 100
                percentages[category] = percentage
            for category, percent in percentages.items():
                expected = 0 if "_PRE" in category else 100
                print(f"{category}: {percent:.2f}% (expected {expected}%)")
            results[sample.title] = test_results
        except Exception as e:
            print(f"Error: {e}")
            results[sample.title] = {"error": str(e)}
        finally:
            with open(SWE_VALIDATION_RESULTS_PATH, "w") as f:
                json.dump(results, f, indent=4)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("targets", nargs="*")
    parser.add_argument("--refresh", "-r", action="store_true")
    parsed_args = parser.parse_args()

    validate_swe_samples(targets=[str(arg) for arg in parsed_args.targets], refresh=parsed_args.refresh)



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Set, TypedDict, Union

from ragdaemon.daemon import Daemon

from mentat.code_feature import CodeFeature, get_consolidated_feature_refs
from mentat.diff_context import DiffContext
from mentat.errors import PathValidationError
from mentat.include_files import (
    PathType,
    get_code_features_for_path,
    get_path_type,
    match_path_with_patterns,
    validate_and_format_path,
)
from mentat.interval import parse_intervals, split_intervals_from_path
from mentat.llm_api_handler import get_max_tokens
from mentat.session_context import SESSION_CONTEXT
from mentat.session_stream import SessionStream
from mentat.utils import get_relative_path, mentat_dir_path


class ContextStreamMessage(TypedDict):
    cwd: str
    diff_context_display: Optional[str]
    auto_context_tokens: int
    features: List[str]
    git_diff_paths: List[str]
    git_untracked_paths: List[str]
    total_tokens: int
    maximum_tokens: int
    total_cost: float


graphs_dir = mentat_dir_path / "ragdaemon"
graphs_dir.mkdir(parents=True, exist_ok=True)


class CodeContext:
    daemon: Daemon

    def __init__(
        self,
        stream: SessionStream,
        cwd: Path,
        diff: Optional[str] = None,
        pr_diff: Optional[str] = None,
        ignore_patterns: Iterable[Path | str] = [],
    ):
        self.diff = diff
        self.pr_diff = pr_diff
        self.ignore_patterns = set(Path(p) for p in ignore_patterns)

        self.diff_context = DiffContext(stream, cwd, self.diff, self.pr_diff)

        self.include_files: Dict[Path, List[CodeFeature]] = {}
        self.ignore_files: Set[Path] = set()

    async def refresh_daemon(self):
        """Call before interacting with context to ensure daemon is up to date."""

        if not hasattr(self, "daemon"):
            # Daemon is initialized after setup because it needs the embedding_provider.
            ctx = SESSION_CONTEXT.get()
            cwd = ctx.cwd
            llm_api_handler = ctx.llm_api_handler

            annotators: dict[str, dict[str, Any]] = {
                "hierarchy": {"ignore_patterns": [str(p) for p in self.ignore_patterns]},
                "chunker_line": {"lines_per_chunk": 50},
                "diff": {"diff": self.diff_context.target},
            }
            self.daemon = Daemon(
                cwd=cwd,
                annotators=annotators,
                verbose=False,
                graph_path=graphs_dir / f"ragdaemon-{cwd.name}.json",
                spice_client=llm_api_handler.spice,
                model=ctx.config.embedding_model,
                provider=ctx.config.embedding_provider,
            )
        await self.daemon.update()

    async def refresh_context_display(self):
        """
        Sends a message to the client with the code context. It is called in the main loop.
        """
        ctx = SESSION_CONTEXT.get()

        diff_context_display = self.diff_context.get_display_context()

        features = get_consolidated_feature_refs(
            [feature for file_features in self.include_files.values() for feature in file_features]
        )
        git_diff_paths = [str(p) for p in self.diff_context.diff_files()]
        git_untracked_paths = [str(p) for p in self.diff_context.untracked_files()]

        total_tokens = await ctx.conversation.count_tokens(include_code_message=True)

        total_cost = ctx.llm_api_handler.spice.total_cost

        data = ContextStreamMessage(
            cwd=str(ctx.cwd),
            diff_context_display=diff_context_display,
            auto_context_tokens=ctx.config.auto_context_tokens,
            features=features,
            git_diff_paths=git_diff_paths,
            git_untracked_paths=git_untracked_paths,
            total_tokens=total_tokens,
            maximum_tokens=get_max_tokens(),
            total_cost=total_cost,
        )
        ctx.stream.send(data, channel="context_update")

    async def get_code_message(
        self,
        prompt_tokens: int,
        prompt: Optional[str] = None,
        expected_edits: Optional[list[str]] = None,  # for training/benchmarking
    ) -> str:
        """
        Retrieves the current code message.
        'prompt' argument is embedded and used to search for similar files when auto-context is enabled.
        If prompt is empty, auto context won't be used.
        'prompt_tokens' argument is the total number of tokens used by the prompt before the code message,
        used to ensure that the code message won't overflow the model's context size
        """
        session_context = SESSION_CONTEXT.get()
        config = session_context.config
        llm_api_handler = session_context.llm_api_handler
        model = config.model
        cwd = session_context.cwd
        code_file_manager = session_context.code_file_manager

        # Setup the header (Mentat-specific, before ragdaemon context)
        header_lines = list[str]()
        self.diff_context.refresh()
        if self.diff_context.diff_files():
            header_lines += [f"Diff References: {self.diff_context.name}\n"]
        header_lines += ["Code Files:\n\n"]

        # Setup a ContextBuilder from Mentat's include_files / diff_context
        await self.refresh_daemon()
        context_builder = self.daemon.get_context("", max_tokens=0)
        diff_nodes: list[str] = [
            node
            for node, data in self.daemon.graph.nodes(data=True)  # pyright: ignore
            if data and "type" in data and data["type"] == "diff"
        ]
        if not self.include_files.values():
            for node in diff_nodes:
                context_builder.add_diff(node)
        for path, features in self.include_files.items():
            for feature in features:
                interval_string = feature.interval_string()
                if interval_string and "-" in interval_string:
                    start, exclusive_end = interval_string.split("-")
                    inclusive_end = str(int(exclusive_end) - 1)
                    interval_string = f"{start}-{inclusive_end}"
                ref = feature.rel_path(session_context.cwd) + interval_string
                context_builder.add_ref(ref, tags=["user-included"])
            relative_path = get_relative_path(path, cwd).as_posix()
            diffs_for_path = [node for node in diff_nodes if f":{relative_path}" in node]
            for diff in diffs_for_path:
                context_builder.add_diff(diff)

        # If auto-context, replace the context_builder with a new one
        if config.auto_context_tokens > 0 and prompt:
            meta_tokens = llm_api_handler.spice.count_tokens("\n".join(header_lines), model, is_message=True)

            include_files_message = context_builder.render()
            include_files_tokens = llm_api_handler.spice.count_tokens(include_files_message, model, is_message=False)

            tokens_used = prompt_tokens + meta_tokens + include_files_tokens
            auto_tokens = min(
                get_max_tokens() - tokens_used - config.token_buffer,
                config.auto_context_tokens,
            )
            context_builder = self.daemon.get_context(
                query=prompt,
                context_builder=context_builder,  # Pass include_files / diff_context to ragdaemon
                max_tokens=get_max_tokens(),
                auto_tokens=auto_tokens,
            )
            for ref in context_builder.to_refs():
                path, interval_str = split_intervals_from_path(Path(ref))
                intervals = parse_intervals(interval_str)
                for interval in intervals:
                    feature = CodeFeature(cwd / path, interval)
                    self.include_features([feature])  # Save ragdaemon context back to include_files

        # The context message is rendered by ragdaemon (ContextBuilder.render())
        context_message = context_builder.render()
        for relative_path in context_builder.context.keys():
            path = Path(cwd / relative_path).resolve()
            if path not in code_file_manager.file_lines:
                with open(path, "r") as file:  # Used by code_file_manager to validate file_edits
                    lines = file.read().split("\n")
                    code_file_manager.file_lines[path] = lines
        return "\n".join(header_lines) + context_message

    def get_all_features(
        self,
        max_chars: int = 100000,
        split_intervals: bool = True,
    ) -> list[CodeFeature]:
        """
        Retrieves every CodeFeature under the cwd. If files_only is True the features won't be split into intervals
        """
        session_context = SESSION_CONTEXT.get()
        cwd = session_context.cwd

        all_features = list[CodeFeature]()
        for _, data in self.daemon.graph.nodes(data=True):  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
            if data is None or "type" not in data or "ref" not in data or data["type"] not in {"file", "chunk"}:  # pyright: ignore[reportUnnecessaryComparison]
                continue
            path, interval = split_intervals_from_path(data["ref"])
            intervals = parse_intervals(interval)
            if not intervals:
                all_features.append(CodeFeature(cwd / path))
            for _interval in intervals:
                all_features.append(CodeFeature(cwd / path, _interval))
        return all_features

    def include_features(self, code_features: Iterable[CodeFeature]):
        """
        Adds the given code features to context. If the feature is already included, it will not be added.
        """
        included_paths: Set[Path] = set()
        for code_feature in code_features:
            if code_feature.path not in self.include_files:
                self.include_files[code_feature.path] = [code_feature]
                included_paths.add(Path(str(code_feature)))
            else:
                code_feature_not_included = True
                for included_code_feature in self.include_files[code_feature.path]:
                    # Intervals can still overlap if user includes intervals different than what chunker breaks up,
                    # but we merge when making code message and don't duplicate lines
                    if (
                        included_code_feature.interval == code_feature.interval
                        # No need to include an interval if the entire file is already included
                        or included_code_feature.interval.whole_file()
                    ):
                        code_feature_not_included = False
                        break
                if code_feature_not_included:
                    if code_feature.interval.whole_file():
                        self.include_files[code_feature.path] = []
                    self.include_files[code_feature.path].append(code_feature)
                    included_paths.add(Path(str(code_feature)))
        return included_paths

    def include(self, path: Path | str, exclude_patterns: Iterable[Path | str] = []) -> Set[Path]:
        """
        Add paths to the context

        Paths that are already in the context and invalid paths are ignored.

        The following are behaviors for each `path` type:

        - File: the file is added to the context.
        - File interval: the file is added to the context, and the interval is added to the file
        - Directory: all files in the directory are recursively added to the context
        - Glob pattern: all files that match the glob pattern are added to the context

        Args:
            `path`: can be a relative or absolute file path, file interval path, directory, or glob pattern.

        Return:
            A set of paths that have been successfully included in the context
        """
        session_context = SESSION_CONTEXT.get()

        path = Path(path)

        abs_exclude_patterns: Set[Path] = set()
        all_exclude_patterns: Set[Union[str, Path]] = set(
            [
                *exclude_patterns,
                *self.ignore_patterns,
                *session_context.config.file_exclude_glob_list,
            ]
        )
        for pattern in all_exclude_patterns:
            if not Path(pattern).is_absolute():
                abs_exclude_patterns.add(session_context.cwd / pattern)
            else:
                abs_exclude_patterns.add(Path(pattern))

        try:
            code_features = get_code_features_for_path(
                path=path,
                cwd=session_context.cwd,
                exclude_patterns=abs_exclude_patterns,
            )
        except PathValidationError as e:
            session_context.stream.send(str(e), style="error")
            return set()

        return self.include_features(code_features)

    def _exclude_file(self, path: Path) -> Path | None:
        session_context = SESSION_CONTEXT.get()
        if path in self.include_files:
            del self.include_files[path]
            return path
        else:
            session_context.stream.send(f"Path {path} not in context", style="error")

    def _exclude_file_interval(self, path: Path) -> Set[Path]:
        session_context = SESSION_CONTEXT.get()

        excluded_paths: Set[Path] = set()

        interval_path, interval_str = split_intervals_from_path(path)
        if interval_path not in self.include_files:
            session_context.stream.send(f"Path {interval_path} not in context", style="error")
            return excluded_paths

        intervals = parse_intervals(interval_str)
        included_code_features: List[CodeFeature] = []
        for code_feature in self.include_files[interval_path]:
            if code_feature.interval not in intervals:
                included_code_features.append(code_feature)
            else:
                excluded_paths.add(Path(str(code_feature)))

        if len(included_code_features) == 0:
            del self.include_files[interval_path]
        else:
            self.include_files[interval_path] = included_code_features

        return excluded_paths

    def _exclude_directory(self, path: Path) -> Set[Path]:
        excluded_paths: Set[Path] = set()

        paths_to_exclude: Set[Path] = set()
        for included_path in self.include_files:
            if path in included_path.parents:
                paths_to_exclude.add(included_path)
        for excluded_path in paths_to_exclude:
            del self.include_files[excluded_path]
            excluded_paths.add(excluded_path)

        return excluded_paths

    def _exclude_glob(self, path: Path) -> Set[Path]:
        excluded_paths: Set[Path] = set()

        paths_to_exclude: Set[Path] = set()
        for included_path in self.include_files:
            if match_path_with_patterns(included_path, set([path])):
                paths_to_exclude.add(included_path)
        for excluded_path in paths_to_exclude:
            del self.include_files[excluded_path]
            excluded_paths.add(excluded_path)

        return excluded_paths

    def exclude(self, path: Path | str) -> Set[Path]:
        """Remove code from the context

        Paths that are not in the context and invalid paths are ignored.

        The following are behaviors for each `path` type:

        - File: the file is removed from the context
        - File interval: the interval is removed from the file. If the file doesn't have the exact interval specified,
          nothing is removed.
        - Directory: all files in the directory are removed from the context
        - Glob pattern: all files that match the glob pattern are removed from the context

        Args:
            `path`: can be a relative or absolute file path, file interval path, directory, or glob pattern.

        Return:
            A set of paths that have been successfully excluded from the context
        """
        session_context = SESSION_CONTEXT.get()

        path = Path(path)
        excluded_paths: Set[Path] = set()
        try:
            validated_path = validate_and_format_path(path, session_context.cwd, check_for_text=False)
            match get_path_type(validated_path):
                case PathType.FILE:
                    excluded_path = self._exclude_file(validated_path)
                    if excluded_path:
                        excluded_paths.add(excluded_path)
                case PathType.FILE_INTERVAL:
                    excluded_paths.update(self._exclude_file_interval(validated_path))
                case PathType.DIRECTORY:
                    excluded_paths.update(self._exclude_directory(validated_path))
                case PathType.GLOB:
                    excluded_paths.update(self._exclude_glob(validated_path))
        except PathValidationError as e:
            session_context.stream.send(str(e), style="error")
            pass

        return excluded_paths

    async def search(
        self,
        query: str,
        max_results: int | None = None,
    ) -> list[tuple[CodeFeature, float]]:
        """Return the top n features that are most similar to the query."""

        cwd = SESSION_CONTEXT.get().cwd
        all_nodes_sorted = self.daemon.search(query, max_results)
        all_features_sorted = list[tuple[CodeFeature, float]]()
        for node in all_nodes_sorted:
            if node.get("type") not in {"file", "chunk"}:
                continue
            distance = node["distance"]
            path, interval = split_intervals_from_path(Path(node["ref"]))
            intervals = parse_intervals(interval)
            for _interval in intervals:
                feature = CodeFeature(cwd / path, _interval)
                all_features_sorted.append((feature, distance))
        if max_results is None:
            return all_features_sorted
        else:
            return all_features_sorted[:max_results]

    def to_simple_context_dict(self) -> dict[str, list[str]]:
        """Return a simple dictionary representation of the code context"""

        simple_dict: dict[str, list[str]] = {}
        for path, features in self.include_files.items():
            simple_dict[str(path.absolute())] = [str(feature) for feature in features]
        return simple_dict

    def from_simple_context_dict(self, simple_dict: dict[str, list[str]]):
        """Load the code context from a simple dictionary representation"""

        for path_str, features_str in simple_dict.items():
            path = Path(path_str)
            features_for_path: List[CodeFeature] = []

            for feature_str in features_str:
                feature_path = Path(feature_str)

                # feature_path is already absolute, so cwd doesn't matter
                current_features = get_code_features_for_path(feature_path, cwd=Path("/"))
                features_for_path += list(current_features)

            self.include_files[path] = features_for_path



#!/usr/bin/env python
import asyncio
import multiprocessing
import os
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import List

import tqdm
from openai import BadRequestError
from spice import SpiceMessage

from benchmarks.arg_parser import common_benchmark_parser
from benchmarks.benchmark_result import BenchmarkResult
from benchmarks.benchmark_run import BenchmarkRun
from benchmarks.exercise_runners.exercise_runner_factory import ExerciseRunnerFactory
from mentat import Mentat
from mentat.config import Config
from mentat.git_handler import get_mentat_branch, get_mentat_hexsha
from mentat.sampler.utils import clone_repo
from mentat.session_context import SESSION_CONTEXT


def clone_exercism_repo(refresh_repo, language):
    exercism_url = f"https://github.com/exercism/{language}.git"
    local_dir = clone_repo(exercism_url, f"exercism-{language}", refresh_repo)
    os.chdir(local_dir)


prompt = (
    "You are a professional code reviewer who helps other coders improve their skills."
    " You recently assigned a coder a small coding test to assess their level, with a"
    " pre-written stub template and an automated test suite to determine if they"
    " succeeded. They failed the test; using the instructions for the test, the code"
    " they wrote for the test, and the output of the test suite, your job is to"
    " determine why they failed. Give a terse but informative couple of sentences on"
    " why they failed, before giving the reason it failed on the final line in the"
    " format:\n"
    + "reason: <reason_failed>\n"
    + "Your response will be parsed programmatically, so you MUST follow the format for"
    " the final line! The possible responses for the final line and what they mean"
    " are as follows:\n" + "blank (the coder didn't change the file at all from the stub you provided"
    " them)\n" + "wording (everything was correct, but the coder messed up the wording or spacing"
    " which caused it to be rejected)\n"
    + "duplication (the coder had a random duplicated line that caused the code to not"
    " be compiled/interpreted)\n" + "syntax (the coder messed up their syntax, meaning their code couldn't be"
    " compiled/interpreted)\n"
    + "logic (the coder messed up the logic)\n"
    + "other (some other reason caused it to fail)\n"
)
model = "gpt-4-0314"


async def failure_analysis(exercise_runner, language):
    instructions = exercise_runner.read_instructions()
    code = exercise_runner.read_code(language)
    test_results = exercise_runner.read_test_results()

    final_message = f"All instructions:\n{instructions}\nCode to review:\n{code}\nTest" f" results:\n{test_results}"
    messages: List[SpiceMessage] = [
        {"role": "system", "content": prompt},
        {"role": "user", "content": final_message},
    ]
    response = ""
    try:
        llm_api_handler = SESSION_CONTEXT.get().llm_api_handler
        llm_grade = await llm_api_handler.call_llm_api(messages, model, None, False)
        response = llm_grade.text
    except BadRequestError:
        response = "Unable to analyze test case\nreason: too many tokens to analyze"

    try:
        reason = response.strip().split("\n")[-1][8:]
    except IndexError:
        reason = "response error"
    return response, reason


async def run_exercise(problem_dir, language="python", max_iterations=2):
    exercise_runner = ExerciseRunnerFactory.create(language, problem_dir)
    client = Mentat(
        cwd=Path("."),
        paths=exercise_runner.include_files(),
        exclude_paths=exercise_runner.exclude_files(),
        config=Config(),
    )
    await client.startup()

    # These prompts are copied from Aider: https://aider.chat/docs/benchmarks.html to
    # allow for direct comparison.
    prompt_1 = (
        f"Use the instructions in {exercise_runner.docs()} to modify"
        + f" {exercise_runner.file}. Keep and implement the existing"
        + " function or class stubs, they will be called from unit tests. Only use"
        + f" standard {language} libraries, don't suggest installing any packages."
    )
    prompt_2 = (
        "\nSee the testing errors above. The tests are correct. Fix the code"
        + f" in {exercise_runner.file} to resolve the errors."
    )

    iterations = 0
    while iterations < max_iterations and not client.stopped.is_set():
        if exercise_runner.passed():
            break
        message = prompt_1 if iterations == 0 else exercise_runner.get_error_message() + prompt_2
        await client.call_mentat_auto_accept(message)

        exercise_runner.run_test()
        iterations += 1

    had_error = client.stopped.is_set()
    messages = client.get_conversation().literal_messages
    await client.shutdown()
    passed = exercise_runner.passed()
    result = BenchmarkResult(
        iterations=iterations,
        passed=passed,
        name=exercise_runner.name,
        tokens=None,
        cost=SESSION_CONTEXT.get().llm_api_handler.spice.total_cost / 100,
        transcript={"id": problem_dir, "messages": messages},
    )
    if had_error:
        result.response = "Error while running mentat"
        result.reason = "error"
    elif not result.passed:
        response, reason = await failure_analysis(exercise_runner, language)
        result.response = response
        result.reason = reason
    return result


def run_exercise_sync(problem_dir, language="python", max_iterations=2):
    exercise_runner = ExerciseRunnerFactory.create(language, problem_dir)
    old_result = exercise_runner.get_result_from_txt()
    if old_result:
        return old_result
    try:
        result = asyncio.run(run_exercise(problem_dir, language, max_iterations))
    except Exception as e:
        print(f"\nError running {problem_dir}")
        print(str(e), flush=True)
        result = BenchmarkResult(
            iterations=0,
            passed=False,
            name=problem_dir,
            tokens=0,
            cost=0,
            response=str(e),
            reason="error",
            transcript={"id": problem_dir, "messages": []},
        )
    result.instructions = exercise_runner.read_instructions()
    result.code = exercise_runner.read_code(language)
    result.test_output = exercise_runner.read_test_results()
    with open("results.txt", "a") as f:
        f.write(result.to_json())
        f.write("\n")
    return result


def tqdm_summary(results):
    passed_in_n = {}
    failed = 0
    for result in results:
        if result.passed:
            iteration = result.iterations
            if iteration:
                passed_in_n[iteration] = passed_in_n.get(iteration, 0) + 1
        else:
            failed += 1
    return "Passed: " + str(passed_in_n)[1:-1] + "| Failed: " + str(failed)


def run_exercism_benchmark(
    benchmarks,
    max_benchmarks,
    max_iterations,
    max_workers,
    language,
):
    all_exercises = os.listdir("exercises/practice")
    if len(benchmarks) > 0:
        exercises = list(set(benchmarks) & set(all_exercises))
    else:
        exercises = all_exercises[:max_benchmarks]
    exercises = sorted(exercises)
    num_exercises = len(exercises)

    # TODO: aiomultiprocessing would be faster with fewer workers; setup a Manager in a parent process
    # that controls the children processes so that we don't run into rate limits
    with multiprocessing.Pool(processes=max_workers) as pool:
        pbar = tqdm.tqdm(total=num_exercises)

        result_map = pool.imap(
            partial(run_exercise_sync, language=language, max_iterations=max_iterations),
            exercises,
        )
        results = []
        for result in result_map:
            pbar.update()
            results.append(result)
            pbar.set_description(tqdm_summary(results) + "| Last Ran: " + result.name)
        results.sort(key=lambda result: result.name)

        benchmark_run = BenchmarkRun(
            results,
            metadata={
                "language": language,
                "type": "exercism",
                "date": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                "commit": get_mentat_hexsha(),
                "branch": get_mentat_branch(),
            },
        )
        benchmark_run.save()
        benchmark_run.render_results()


if __name__ == "__main__":
    parser = common_benchmark_parser()
    args = parser.parse_args()
    clone_exercism_repo(args.refresh_repo, args.language)
    print(args)
    run_exercism_benchmark(
        args.benchmarks,
        args.max_benchmarks,
        args.max_iterations,
        args.max_workers,
        args.language,
    )



from __future__ import annotations

import json
from argparse import ArgumentParser, Namespace
from json import JSONDecodeError
from pathlib import Path
from typing import Optional

import attr
from attr import converters, validators
from spice.models import TextModel, models
from spice.spice import EmbeddingModel

from mentat.git_handler import get_git_root_for_path
from mentat.parsers.parser import Parser
from mentat.parsers.parser_map import parser_map
from mentat.session_context import SESSION_CONTEXT
from mentat.terminal.themes import themes
from mentat.utils import mentat_dir_path

config_file_name = Path(".mentat_config.json")
user_config_path = mentat_dir_path / config_file_name


def int_or_none(s: str | None) -> int | None:
    if s is not None:
        return int(s)
    return None


bool_autocomplete = ["True", "False"]


@attr.define
class Config:
    _errors: list[str] = attr.field(factory=list)

    # Model specific settings
    model: str = attr.field(
        default="gpt-4-0125-preview",
        metadata={"auto_completions": [model.name for model in models if isinstance(model, TextModel)]},
    )
    provider: Optional[str] = attr.field(default=None, metadata={"auto_completions": ["openai", "anthropic", "azure"]})
    embedding_model: str = attr.field(
        default="text-embedding-ada-002",
        metadata={"auto_completions": [model.name for model in models if isinstance(model, EmbeddingModel)]},
    )
    embedding_provider: Optional[str] = attr.field(
        default=None, metadata={"auto_completions": ["openai", "anthropic", "azure"]}
    )
    temperature: float = attr.field(default=0.2, converter=float, validator=[validators.le(1), validators.ge(0)])

    maximum_context: int | None = attr.field(
        default=None,
        metadata={
            "description": (
                "The maximum number of lines of context to include in the prompt. It is"
                " inferred automatically for openai models but you can still set it to"
                " save costs. If not set for non-openai models, it defaults to 4096."
            ),
        },
        converter=int_or_none,
        validator=validators.optional(validators.ge(0)),
    )
    token_buffer: int = attr.field(
        default=1000,
        metadata={
            "description": ("The amount of tokens to always be reserved as a buffer for user and model messages."),
        },
    )
    parser: Parser = attr.field(  # pyright: ignore
        default="block",
        metadata={
            "description": (
                "The format for the LLM to write code in. You probably don't want to mess with this setting."
            ),
            "auto_completions": list(parser_map.keys()),
        },
        converter=parser_map.get,  # pyright: ignore
        validator=validators.instance_of(Parser),  # pyright: ignore
    )
    no_parser_prompt: bool = attr.field(
        default=False,
        metadata={
            "description": (
                "Whether to include the parser prompt in the system message. This"
                " should only be set to true for fine tuned models."
            ),
            "auto_completions": bool_autocomplete,
        },
        converter=converters.optional(converters.to_bool),
    )
    revisor: bool = attr.field(
        default=False,
        metadata={
            "description": (
                "Enables or disables a revisor tweaking model edits after they're made."
                " The revisor will use the same model regular edits do."
            ),
            "auto_completions": bool_autocomplete,
        },
        converter=converters.optional(converters.to_bool),
    )
    sampler: bool = attr.field(
        default=False,
        metadata={
            "description": ("Automatically saves a git diff snapshot for the sampler on startup."),
            "auto_completions": bool_autocomplete,
        },
        converter=converters.optional(converters.to_bool),
    )

    # Context specific settings
    file_exclude_glob_list: list[str] = attr.field(
        factory=list,
        metadata={"description": "List of glob patterns to exclude from context"},
    )
    auto_context_tokens: int = attr.field(  # pyright: ignore
        default=0,
        metadata={
            "description": (
                "Automatically selects code files for every request to include in"
                " context. Adds this many tokens to context each request."
            ),
            "abbreviation": "a",
            "const": 5000,
        },
        converter=int,
        validator=validators.ge(0),  # pyright: ignore
    )

    # Sample specific settings
    sample_repo: str | None = attr.field(
        default=None,
        metadata={"description": "A public url for a cloneable git repository to sample from."},
    )
    sample_merge_base_target: str | None = attr.field(
        default=None,
        metadata={"description": "The branch or commit to use as the merge base for samples."},
    )

    theme: str = attr.field(  # pyright: ignore
        default="dark",
        metadata={
            "description": "Theme for interaction possible choices are light or dark.",
            "no_midsession_change": True,
        },
        validator=validators.in_(themes.keys()),  # pyright: ignore
    )

    @classmethod
    def get_fields(cls) -> list[str]:
        return [field.name for field in attr.fields(cls) if not field.name.startswith("_")]

    @classmethod
    def add_fields_to_argparse(cls, parser: ArgumentParser) -> None:
        for field in attr.fields(cls):
            if "no_flag" in field.metadata:
                continue
            name = [f"--{field.name.replace('_', '-')}"]
            if "abbreviation" in field.metadata:
                name.append(f"-{field.metadata['abbreviation'].replace('_', '-')}")

            arguments = {
                "help": field.metadata.get("description", ""),
            }
            if "const" in field.metadata:
                arguments["nargs"] = "?"
                arguments["const"] = field.metadata["const"]

            if field.type == "bool":
                if arguments.get("default", False):
                    arguments["action"] = "store_false"
                else:
                    arguments["action"] = "store_true"
            elif field.type == "int":
                arguments["type"] = int
            elif field.type == "float":
                arguments["type"] = float
            elif field.type == "list[str]":
                arguments["nargs"] = "*"

            parser.add_argument(*name, **arguments)

    @classmethod
    def create(cls, cwd: Path, args: Namespace | None = None) -> Config:
        config = Config()

        # Each method overwrites the previous so they are in order of precedence
        config.load_file(user_config_path)
        git_root = get_git_root_for_path(cwd, raise_error=False)
        if git_root is not None:
            config.load_file(git_root / config_file_name)
        config.load_file(cwd / config_file_name)

        if args is not None:
            config.load_namespace(args)

        return config

    def load_namespace(self, args: Namespace) -> None:
        for field in attr.fields(Config):
            if field.name in args and field.name != "_errors":
                value = getattr(args, field.name)
                if value is not None:
                    try:
                        setattr(self, field.name, value)
                    except (ValueError, TypeError) as e:
                        self.error(f"Warning: Illegal value for {field}: {e}")

    def load_file(self, path: Path) -> None:
        if path.exists():
            with open(path) as config_file:
                try:
                    config = json.load(config_file)
                except JSONDecodeError:
                    self.error(f"Warning: Config {path} contains invalid json; ignoring user configuration file")
                    return
            for field in config:
                if hasattr(self, field):
                    try:
                        setattr(self, field, config[field])
                    except (ValueError, TypeError) as e:
                        self.error(f"Warning: Config {path} contains invalid value for" f" setting: {field}\n{e}")
                else:
                    self.error(f"Warning: Config {path} contains unrecognized setting: {field}")

    def error(self, message: str) -> None:
        self._errors.append(message)
        try:
            self.send_errors_to_stream()
        except LookupError:
            pass

    def send_errors_to_stream(self):
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        for error in self._errors:
            stream.send(error, style="warning")
        self._errors = []



from __future__ import annotations

from collections import defaultdict
from pathlib import Path
from typing import Optional

import attr
from ragdaemon.utils import get_document

from mentat.errors import MentatError
from mentat.interval import INTERVAL_FILE_END, Interval
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import get_relative_path

MIN_INTERVAL_LINES = 10


@attr.define(frozen=True)
class CodeFeature:
    """
    Represents a section of the code_message which is included with the prompt.
    Includes a section of code and an annotation method.

    Attributes:
        path: The absolute path to the file.
        interval: The lines in the file.
        name: The names of the features/functions in this CodeFeature
    """

    path: Path = attr.field()
    interval: Interval = attr.field(factory=lambda: Interval(1, INTERVAL_FILE_END))
    # eq is set to false so that we can compare duplicate features without names getting in the way
    name: Optional[str] = attr.field(default=None, eq=False)

    def __attrs_post_init__(self):
        if not self.path.is_absolute():
            raise MentatError("CodeFeature path must be absolute.")

    def __repr__(self):
        return (
            f"CodeFeature(path={self.path}," f" interval={self.interval.start}-{self.interval.end}, name={self.name})"
        )

    def rel_path(self, cwd: Optional[Path] = None) -> str:
        if cwd is not None:
            path_string = str(get_relative_path(self.path, cwd))
        else:
            path_string = str(self.path)
        return path_string

    def interval_string(self) -> str:
        if not self.interval.whole_file():
            interval_string = f":{self.interval.start}-{self.interval.end}"
        else:
            interval_string = ""
        return interval_string

    def __str__(self, cwd: Optional[Path] = None) -> str:
        return self.rel_path(cwd) + self.interval_string()


def count_feature_tokens(feature: CodeFeature, model: str) -> int:
    ctx = SESSION_CONTEXT.get()

    cwd = ctx.cwd
    ref = feature.__str__(cwd)
    document = get_document(ref, cwd)
    return ctx.llm_api_handler.spice.count_tokens(document, model, is_message=False)


def get_consolidated_feature_refs(features: list[CodeFeature]) -> list[str]:
    """
    Return a list of 'path:<interval>,<interval>' strings, merging code features with the same path
    """
    level_info_by_path = defaultdict[Path, list[Interval | None]](list)
    for f in features:
        if f.interval.whole_file():
            level_info_by_path[f.path].append(None)
        else:
            level_info_by_path[f.path].append(f.interval)

    consolidated_refs = list[str]()
    for path, level_info in level_info_by_path.items():
        ref_string = str(path)
        intervals = sorted([level for level in level_info if level is not None])
        if intervals and None not in level_info:
            ref_string += ":" + ",".join(str(interval) for interval in intervals)
        consolidated_refs.append(ref_string)

    return consolidated_refs



import subprocess
from pathlib import Path
from typing import List, Literal, Optional

from mentat.git_handler import (
    check_head_exists,
    get_diff_for_file,
    get_files_in_diff,
    get_git_root_for_path,
    get_treeish_metadata,
    get_untracked_files,
)
from mentat.session_context import SESSION_CONTEXT
from mentat.session_stream import SessionStream


class DiffContext:
    target: str = ""
    name: str = "index (last commit)"

    def __init__(
        self,
        stream: SessionStream,
        cwd: Path,
        diff: Optional[str] = None,
        pr_diff: Optional[str] = None,
    ):
        self.git_root = get_git_root_for_path(cwd, raise_error=False)
        if not self.git_root:
            return

        if diff and pr_diff:
            # TODO: Once broadcast queue's unread messages and/or config is moved to client,
            # determine if this should quit or not
            stream.send(
                "Cannot specify more than one type of diff. Disabling diff and pr-diff.",
                style="warning",
            )
            diff = None
            pr_diff = None

        target = diff or pr_diff
        if not target:
            return

        name = ""
        treeish_type = _get_treeish_type(self.git_root, target)
        if treeish_type is None:
            stream.send(f"Invalid treeish: {target}", style="failure")
            stream.send("Disabling diff and pr-diff.", style="warning")
            return

        if treeish_type == "branch":
            name += f"Branch {target}: "
        elif treeish_type in {"relative"}:
            name += f"{target}: "

        if pr_diff:
            name = f"Merge-base {name}"
            target = _git_command(self.git_root, "merge-base", "HEAD", pr_diff)
            if not target:
                # TODO: Same as above todo
                stream.send(
                    f"Cannot identify merge base between HEAD and {pr_diff}. Disabling pr-diff.",
                    style="warning",
                )
                return

        def _get_treeish_metadata(git_root: Path, _target: str):
            meta = get_treeish_metadata(git_root, _target)
            return f'{meta["hexsha"][:8]}: {meta["summary"]}'

        if not target:
            return
        elif treeish_type == "compare":
            name += "Comparing " + ", ".join(_get_treeish_metadata(self.git_root, part) for part in target.split(" "))
        else:
            name += _get_treeish_metadata(self.git_root, target)
        self.target = target
        self.name = name

    _diff_files: List[Path] | None = None
    _untracked_files: List[Path] | None = None

    def diff_files(self) -> List[Path]:
        if not self.git_root:
            return []
        if self._diff_files is None:
            self.refresh()
        return self._diff_files  # pyright: ignore

    def untracked_files(self) -> List[Path]:
        if not self.git_root:
            return []
        if self._untracked_files is None:
            self.refresh()
        return self._untracked_files  # pyright: ignore

    def refresh(self):
        if not self.git_root:
            return
        ctx = SESSION_CONTEXT.get()

        if self.target == "HEAD" and not check_head_exists():
            self._diff_files = []  # A new repo without any commits
            self._untracked_files = []
        else:
            self._diff_files = [(ctx.cwd / f).resolve() for f in get_files_in_diff(self.target)]
            self._untracked_files = [(ctx.cwd / f).resolve() for f in get_untracked_files(ctx.cwd)]

    def get_display_context(self) -> Optional[str]:
        if not self.git_root:
            return None
        diff_files = self.diff_files()
        if not diff_files:
            return ""
        num_files = len(diff_files)
        num_lines = 0
        for file in diff_files:
            diff = get_diff_for_file(self.target, file)
            diff_lines = diff.splitlines()
            num_lines += len([line for line in diff_lines if line.startswith(("+ ", "- "))])
        return f" {self.name} | {num_files} files | {num_lines} lines"


TreeishType = Literal["commit", "branch", "relative", "compare"]


def _git_command(git_root: Path, *args: str) -> str | None:
    try:
        return subprocess.check_output(["git"] + list(args), cwd=git_root, stderr=subprocess.PIPE, text=True).strip()
    except subprocess.CalledProcessError:
        return None


def _get_treeish_type(git_root: Path, treeish: str) -> TreeishType | None:
    if " " in treeish:
        parts = treeish.split(" ")
        types = [_get_treeish_type(git_root, part) for part in parts]
        if not all(types):
            return None
        return "compare"

    object_type = _git_command(git_root, "cat-file", "-t", treeish)

    if not object_type:
        return None

    if object_type == "commit":
        if "~" in treeish or "^" in treeish:
            return "relative"

        if _git_command(git_root, "show-ref", "--heads", treeish):
            return "branch"
        else:
            return "commit"
    return None



from __future__ import annotations

import logging
import os
import sys
from inspect import iscoroutinefunction
from pathlib import Path
from typing import (
    Any,
    Callable,
    List,
    Literal,
    Optional,
    TypeVar,
    overload,
)

import sentry_sdk
from dotenv import load_dotenv
from openai.types.chat.completion_create_params import ResponseFormat
from spice import EmbeddingResponse, Spice, SpiceMessage, SpiceResponse, StreamingSpiceResponse, TranscriptionResponse
from spice.errors import APIConnectionError, NoAPIKeyError
from spice.models import WHISPER_1
from spice.providers import OPEN_AI
from spice.spice import UnknownModelError, get_model_from_name

from mentat.errors import MentatError, ReturnToUser
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import mentat_dir_path

TOKEN_COUNT_WARNING = 32000


def is_test_environment():
    """Returns True if in pytest and not benchmarks"""
    benchmarks_running = os.getenv("MENTAT_BENCHMARKS_RUNNING")
    return (
        "PYTEST_CURRENT_TEST" in os.environ
        and "--benchmark" not in sys.argv
        and (not bool(benchmarks_running) or benchmarks_running == "false")
    )


RetType = TypeVar("RetType")


def api_guard(func: Callable[..., RetType]) -> Callable[..., RetType]:
    """Decorator that should be used on any function that calls the OpenAI API

    It does two things:
    1. Raises if the function is called in tests (that aren't benchmarks)
    2. Converts APIConnectionErrors to MentatErrors
    """

    if iscoroutinefunction(func):

        async def async_wrapper(*args: Any, **kwargs: Any) -> RetType:
            assert not is_test_environment(), "OpenAI call attempted in non-benchmark test environment!"
            try:
                return await func(*args, **kwargs)
            except APIConnectionError:
                raise MentatError("API connection error: please check your internet connection and try again.")
            except UnknownModelError:
                SESSION_CONTEXT.get().stream.send(
                    "Unknown model. Use /config provider <provider> and try again.", style="error"
                )
                raise ReturnToUser()

        return async_wrapper  # pyright: ignore[reportReturnType]
    else:

        def sync_wrapper(*args: Any, **kwargs: Any) -> RetType:
            assert not is_test_environment(), "OpenAI call attempted in non-benchmark test environment!"
            try:
                return func(*args, **kwargs)
            except APIConnectionError:
                raise MentatError("API connection error: please check your internet connection and try again.")
            except UnknownModelError:
                SESSION_CONTEXT.get().stream.send(
                    "Unknown model. Use /config provider <provider> and try again.", style="error"
                )
                raise ReturnToUser()

        return sync_wrapper


# Ensures that each chunk will have at most one newline character
def chunk_to_lines(content: str) -> list[str]:
    return content.splitlines(keepends=True)


def get_max_tokens() -> int:
    session_context = SESSION_CONTEXT.get()
    stream = session_context.stream
    config = session_context.config

    context_size = get_model_from_name(config.model).context_length
    maximum_context = config.maximum_context

    if context_size is not None and maximum_context is not None:
        return min(context_size, maximum_context)
    elif context_size is not None:
        return context_size
    elif maximum_context is not None:
        return maximum_context
    else:
        maximum_context = 4096
        # This attr has a converter from str to int
        config.maximum_context = str(maximum_context)
        stream.send(
            f"Context size for {config.model} is not known. Set maximum-context"
            " with `/config maximum_context <value>`. Using a default value of"
            f" {maximum_context}.",
            style="error",
        )
        return maximum_context


def raise_if_context_exceeds_max(tokens: int):
    ctx = SESSION_CONTEXT.get()

    max_tokens = get_max_tokens()
    if max_tokens - tokens < ctx.config.token_buffer:
        ctx.stream.send(
            f"The context size is limited to {max_tokens} tokens and your current"
            f" request uses {tokens} tokens. Please use `/exclude` to remove"
            " some files from context or `/clear` to reset the conversation",
            style="error",
        )
        raise ReturnToUser()


class LlmApiHandler:
    """Used for any functions that require calling the external LLM API"""

    def __init__(self):
        self.spice = Spice()

    async def initialize_client(self):
        ctx = SESSION_CONTEXT.get()

        if not load_dotenv(mentat_dir_path / ".env") and not load_dotenv(ctx.cwd / ".env"):
            load_dotenv()

        try:
            self.spice.load_provider(OPEN_AI)
        except NoAPIKeyError:
            from mentat.session_input import collect_user_input

            ctx.stream.send(
                "No OpenAI api key detected. To avoid entering your api key on startup, create a .env file in"
                " ~/.mentat/.env or in your workspace root.",
                style="warning",
            )
            ctx.stream.send("Enter your api key:", style="info")
            key = (await collect_user_input(log_input=False)).data
            os.environ["OPENAI_API_KEY"] = key

    @overload
    async def call_llm_api(
        self,
        messages: List[SpiceMessage],
        model: str,
        provider: Optional[str],
        stream: Literal[False],
        response_format: ResponseFormat = ResponseFormat(type="text"),
    ) -> SpiceResponse:
        ...

    @overload
    async def call_llm_api(
        self,
        messages: List[SpiceMessage],
        model: str,
        provider: Optional[str],
        stream: Literal[True],
        response_format: ResponseFormat = ResponseFormat(type="text"),
    ) -> StreamingSpiceResponse:
        ...

    @api_guard
    async def call_llm_api(
        self,
        messages: List[SpiceMessage],
        model: str,
        provider: Optional[str],
        stream: bool,
        response_format: ResponseFormat = ResponseFormat(type="text"),
    ) -> SpiceResponse | StreamingSpiceResponse:
        session_context = SESSION_CONTEXT.get()
        config = session_context.config

        # Confirm that model has enough tokens remaining
        tokens = self.spice.count_prompt_tokens(messages, model)
        raise_if_context_exceeds_max(tokens)

        with sentry_sdk.start_span(description="LLM Call") as span:
            span.set_tag("model", model)

            if not stream:
                response = await self.spice.get_response(
                    model=model,
                    provider=provider,
                    messages=messages,
                    temperature=config.temperature,
                    response_format=response_format,
                )
            else:
                response = await self.spice.stream_response(
                    model=model,
                    provider=provider,
                    messages=messages,
                    temperature=config.temperature,
                    response_format=response_format,
                )

        return response

    @api_guard
    def call_embedding_api(self, input_texts: list[str], model: str = "text-embedding-ada-002") -> EmbeddingResponse:
        ctx = SESSION_CONTEXT.get()
        return self.spice.get_embeddings_sync(input_texts, model, provider=ctx.config.embedding_provider)

    @api_guard
    async def call_whisper_api(self, audio_path: Path) -> TranscriptionResponse:
        return await self.spice.get_transcription(audio_path, model=WHISPER_1)

    def display_cost_stats(self, response: SpiceResponse):
        ctx = SESSION_CONTEXT.get()

        display = f"Speed: {response.characters_per_second:.2f} char/s"
        if response.cost is not None:
            display += f" | Cost: ${response.cost / 100:.2f}"

        costs_logger = logging.getLogger("costs")
        costs_logger.info(display)

        ctx.stream.send(display, style="info")
